import 'package:flutter_blue_plus/flutter_blue_plus.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:permission_handler/permission_handler.dart';

class ModuleBleBuiltin {
  static bool scanFilterWearEcg(BluetoothDevice device) =>
      device.localName.startsWith(DEVICE_NAME_WEAR_ECG_12);

  static bool scanFilterCardioClick(BluetoothDevice device) =>
      device.localName.startsWith(DEVICE_NAME_CARDIO_CLICK);

  static bool scanFilterWearCardioT(BluetoothDevice device) =>
      device.localName.startsWith(DEVICE_NAME_WEAR_CARDIO_T);
}

class ModuleBle {
  ModuleBle({
    required UUID this.uuid,
    required this.streams,
  });

  BluetoothDevice? device;

  bool Function(BluetoothDevice device) scanFilter =
      ModuleBleBuiltin.scanFilterWearEcg;

  bool Function(ScanResult scanResult) get _scanResultFilter =>
      (scanResult) => scanFilter(scanResult.device);

  int _sortByRssi(a, b) => b.rssi - a.rssi;

  DateTime? _sessionTime;
  final StreamInterface streams;

  final UUID uuid;
  final points = <Enum, BluetoothCharacteristic>{};

  Future<bool> connectDevice({
    String preferredDeviceId = '',
  }) async {
    /// 0. initialize
    device = null;
    'Device is reset into $device'.log;

    /// 0. permission request
    final permissions = [
      Permission.bluetoothScan,
      Permission.bluetooth,
      Permission.bluetoothConnect,
      Permission.bluetoothAdvertise,
    ];
    await permissions.request();
    'Permissions are request: $permissions'.log;

    /// 1. devices which are connected
    final connectedDevices = await FlutterBluePlus.connectedSystemDevices;
    final connectedFeasibleDevices = connectedDevices.where(scanFilter);
    if (connectedFeasibleDevices.isNotEmpty) {
      device = connectedFeasibleDevices.firstWhere(
        (device) => device.remoteId.toString() == preferredDeviceId,
        orElse: () => connectedFeasibleDevices.first,
      );
      'Connected with device named ${device?.localName}'.log;
      return true;
    }

    /// 2. perform scanning process
    final scan = FlutterBluePlus.startScan(
      timeout: const Duration(seconds: 2),
    );

    return await scan.then(
      (events) async {
        final scanResults = events as List<ScanResult>;
        final feasibleScanResult = scanResults
            .where(_scanResultFilter)
            .toList(growable: false)
          ..sort(_sortByRssi);

        if (feasibleScanResult.isNotEmpty) {
          final result = feasibleScanResult.firstWhere(
            (r) => r.device.remoteId.toString() == preferredDeviceId,
            orElse: () => feasibleScanResult.first,
          );
          device = result.device;
          // if (Platform.isAndroid) {
          // await device?.pair();
          // 'Try to pair with device named ${device?.localName}'.log;
          // }
          await device?.connect();
          'Connected with device named ${device?.localName}'.log;
          return true;
        }
        return false;
      },
    );
  }

  Future<void> disconnectDevice({
    BluetoothDevice? device,
  }) async {
    if (device != null) {
      await device.disconnect();
      'Device named ${device.localName} has been disconnected now'.log;
      return;
    }

    final connected = await FlutterBluePlus.connectedSystemDevices;

    await Future.wait(
      [
        for (final device in connected.where(scanFilter))
          () async {
            await device.disconnect();
            'Device named ${device.localName} has been disconnected now'.log;
          }(),
      ],
    );
    device = null;
  }

  Future<void> registerCharacteristics() async {
    if (device == null) return;

    final services = await device!.discoverServices();
    for (final service in services) {
      final characteristics = service.characteristics;
      for (final characteristic in characteristics) {
        uuid.characteristic.forEach(
          (key, value) {
            final uuid = characteristic.uuid.toString();
            if (value == null) return;
            if (uuid.toLowerCase() == value.toLowerCase()) {
              '${key.name} was registered'.log;
              points[key] = characteristic;
            }
          },
        );
      }
    }

    _sessionTime = DateTime.now();

    streams
      ..connection = _getConnectionStateStream().broadcast
      ..battery = _getBatteryLevelStream().broadcast
      ..tap = _getPhysicalTapStream().broadcast;
  }

  // Future<void> initializeStreams() async {
  //   if (uuid.enableEcg)
  //
  //   if (uuid.enableEcg)
  // }

  Stream<bool?> _getConnectionStateStream() async* {
    final stream = device?.connectionState.broadcast;
    if (stream == null) return;

    var previousEmitTime = DateTime.now();
    BluetoothConnectionState? prev;

    final subscription = stream.listen((event) {
      prev = event;
      previousEmitTime = DateTime(0);
    });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmitTime);
      if (timeAfterLastEmit > const Duration(seconds: 2)) {
        yield prev == BluetoothConnectionState.connected;
        previousEmitTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 2));
    }
    yield null;

    await subscription.cancel();
  }

  Stream<int?> _getBatteryLevelStream() async* {
    final battery = points[Characteristics.BATTERY_LEVEL];
    if (battery == null) return;

    try {
      await battery.setNotifyValue(true);
    } catch (e) {
      '_getBatteryLevelStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await battery.setNotifyValue(true);
      } catch (e) {
        '_getBatteryLevelStream (second try): $e'.log;
      }
    }

    final read = await battery.read();
    var prev = read.isNotEmpty ? read[0] : null;
    yield prev;

    var previousEmitTime = DateTime.now();
    final subscription = battery.lastValueStream.listen((event) {
      final changed = event.isNotEmpty ? event[0] : null;
      prev = changed;
      previousEmitTime = DateTime(0);
    });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmitTime);
      if (timeAfterLastEmit > const Duration(seconds: 5)) {
        yield prev;
        previousEmitTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 5));
    }

    yield null;

    await subscription.cancel();
  }

  Stream<bool?> _getPhysicalTapStream() async* {
    final tap = points[Characteristics.TAP_BUTTON];
    if (tap == null) return;
    try {
      await tap.setNotifyValue(true);
    } catch (e) {
      '_getPhysicalTapStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await tap.setNotifyValue(true);
      } catch (e) {
        '_getPhysicalTapStream (second try): $e'.log;
      }
    }

    final sessionTimeInside = _sessionTime;

    await for (final event in tap.lastValueStream) {
      if (sessionTimeInside != _sessionTime) break;
      if (event.isEmpty) continue;
      yield true;
    }
    yield null;
  }
}
