import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

class ModuleEcgBuiltin {
  static Stream<List<int>> udfRawStream({
    required ModuleBle ble,
  }) async* {
    final measurement = ble.points[Characteristics.MEASUREMENT_ECG];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }
    await for (final event in measurement.lastValueStream) {
      yield event;
    }
  }

  static List<List<double>>? udfConvert({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    final res = List.generate(
      channelCount,
          (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.isEmpty) return null;
    if (event.length < 4) return null;
    final id = event[3];
    if (id != 0x11) return null;

    /// [V6, L1, L2, V2, V3, V4, V5, V6, V1]
    final ordering = [11, 0, 1, 7, 8, 9, 10, 6];
    final skip = 6;
    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < 8; j++) {
        final msb = event[skip + 24 * i + j];
        final mid = event[skip + 24 * i + j + 1];
        final lsb = event[skip + 24 * i + j + 2];
        final value = toDouble(msb, mid, lsb);
        res[ordering[j]][i] = value;
      }
    }
    for (var i = 0; i < sampleCount; i++) {
      final l1 = res[0][i];
      final l2 = res[1][i];
      res[2].add(l3(l1, l2));
      res[3].add(aVR(l1, l2));
      res[4].add(aVL(l1, l2));
      res[5].add(aVF(l1, l2));
    }
    return res;
  }

  static Stream<List<int>> udfLeadStreamWearEcg({
    required ModuleBle ble,
  }) {
    return ble.streams.eventEcg!;
  }

  static Stream<List<int>> udfLeadStreamCardioClick({
    required ModuleBle ble,
  }) async* {
    final leadState = ble.points[Characteristics.LEAD_STATE_ECG];
    if (leadState == null) return;

    try {
      await leadState.setNotifyValue(true);
    } catch (e) {
      '_getLeadStateStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await leadState.setNotifyValue(true);
      } catch (e) {
        '_getLeadStateStream (second try): $e'.log;
      }
    }
    await for (final event in leadState.lastValueStream) {
      yield event;
    }
  }

  static Channels udfFilter({
    required Channels channels,
    required int channelCount,
    required int sampleCount,
    required DataFilter dataFilter,
    required bool isAttached,
  }) {
    final startIndex = channels.l2.length - sampleCount;
    for (var i = startIndex; i < channels.l2.length; i++) {
      final l1 = dataFilter.applyFilter(0, channels.l1[i], isAttached);
      final l2 = dataFilter.applyFilter(1, channels.l2[i], isAttached);

      channels.l1[i] = l1;
      channels.l2[i] = l2;
      channels.l3[i] = l3(l1, l2);

      channels.aVR[i] = aVR(l1, l2);
      channels.aVL[i] = aVL(l1, l2);
      channels.aVF[i] = aVF(l1, l2);

      channels.v1[i] = dataFilter.applyFilter(2, channels.v1[i], isAttached);
      channels.v2[i] = dataFilter.applyFilter(3, channels.v2[i], isAttached);
      channels.v3[i] = dataFilter.applyFilter(4, channels.v3[i], isAttached);
      channels.v4[i] = dataFilter.applyFilter(5, channels.v4[i], isAttached);
      channels.v5[i] = dataFilter.applyFilter(6, channels.v5[i], isAttached);
      channels.v6[i] = dataFilter.applyFilter(7, channels.v6[i], isAttached);
    }
    return channels;
  }

  static LeadState? udfLeadCompute({
    required List<int> event,
  }) {
    if (event.isEmpty) return null;
    if (event.length < 4) return null;
    final id = event[3];
    if (id != 0x11) return null;
    if (event.length < 6) return null;
    final high = event[4];
    final low = event[5];

    final leadState = LeadState.fromBytes(high, low);
    return leadState;
  }
}

class ModuleEcg {
  ModuleEcg({
    required this.ble,
    this.commands = const Commands(),
    this.udfRawStream = ModuleEcgBuiltin.udfRawStream,
    this.udfConvert = ModuleEcgBuiltin.udfConvert,
    this.udfLeadStream = ModuleEcgBuiltin.udfLeadStreamWearEcg,
    this.udfLeadCompute = ModuleEcgBuiltin.udfLeadCompute,
    this.udfFilter = ModuleEcgBuiltin.udfFilter,
  });

  final ModuleBle ble;
  final Commands commands;

  final Stream<List<int>> Function({
  required ModuleBle ble,
  }) udfRawStream;

  final List<List<double>>? Function({
  required List<int> event,
  required int channelCount,
  required int sampleCount,
  }) udfConvert;

  final Stream<List<int>> Function({
  required ModuleBle ble,
  }) udfLeadStream;

  final LeadState? Function({
  required List<int> event,
  }) udfLeadCompute;

  final Channels Function({
  required Channels channels,
  required int channelCount,
  required int sampleCount,
  required DataFilter dataFilter,
  required bool isAttached,
  }) udfFilter;

  int samplingRate = 250;
  int channelCount = 12;
  int sampleCount = 9;
  int filterCount = 8;

  /// [maxBuffer] decides a maximum length of buffered Stream.
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250 by default.
  int maxBuffer = 40 * 250;

  /// [lengthBuffer] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250 by default.
  int lengthBuffer = 30 * 250;

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int displayBuffer = 5 * 250;

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 250 ~/ 2 by default.
  int refreshBuffer = 250 ~/ 2;

  /// [isFiltered] decides whether ecg signal is filtered.
  bool isFiltered = true;

  /// If [isCompressed] is `true`,
  /// ecg signal is decompressed by LZMA algorithm.
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter.initialize(8);

  AdaptiveHR adaptiveHR = AdaptiveHR(0);

  final List<(void Function(Channels, DataFilter)? callback, int period)>
  callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedEcg]
  Channels? lastChannel;

  /// [lastLeadState] is a latest stream value from [ble.streams.leadStateEcg]
  LeadState? lastLeadState;

  bool _isPaused = false;
  bool _isMeasuring = false;

  ///////////////////////////////////////////////////////////////////////

  void initStream() {
    ble.streams
      ..rawEcg = null
      ..lzmaEcg = null
      ..convertedEcg = null
      ..eventEcg = null
      ..leadStateEcg = null
      ..heartRateEcg = null
      ..bufferedEcg = null
      ..bufferedFilterEcg = null
      ..plotEcg = null
      ..plotRTEcg = null;
  }

  Future<void> start() async {
    dataFilter = DataFilter.initialize(filterCount);
    adaptiveHR = AdaptiveHR(samplingRate == 250 ? 0 : 1);

    await _sendDate();
    await _sendStartSave();
    await _sendStart();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams
      ..rawEcg = _getRaw().broadcast
      ..lzmaEcg = _getLzma().broadcast
      ..convertedEcg = _getConverted().broadcast
      ..eventEcg = _getEvent()?.broadcast
      ..leadStateEcg = _getLeadState().broadcast
      ..heartRateEcg = _getHeartRate().broadcast
      ..bufferedEcg = _getBuffered().broadcast
      ..bufferedFilterEcg = _getBufferedFilter().broadcast
      ..plotEcg = _getPlot().broadcast
      ..plotRTEcg = _getPlotRT().broadcast;
  }

  /// [pause]
  Future<void> pause() async {
    _isPaused = !_isPaused;
  }

  /// [stop] sends stop_ecg_packet to device.
  Future<void> stop() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_ECG];
    await control.writePacket(commands.stopEcg);
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_ECG];
    await control.writePacket(commands.startEcg);
  }

  Future<void> _sendDate() async {
    final control = ble.points[Characteristics.SET_DATA];
    await control.writePacket(commands.setDate);
  }

  Future<void> _sendStartSave() async {
    final control = ble.points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.save.startEcg);
  }

  Stream<List<int>> _getRaw() async* {
    final events = udfRawStream(ble: ble);

    await for (final res in events) {
      if (!_isMeasuring || ble.device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.rawEcg;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed ? ble.streams.lzmaEcg : ble.streams.rawEcg;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  ///
  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.eventEcg;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  ///
  Stream<LeadState?> _getLeadState() async* {
    final events = udfLeadStream(ble: ble);
    var now = DateTime.now();

    await for (final event in events) {
      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      final leadState = udfLeadCompute(event: event);
      if (leadState == null) continue;
      yield leadState;

      lastLeadState = leadState;
    }
    yield null;
  }

  ///
  Stream<int?> _getHeartRate() async* {
    final events = ble.streams.convertedEcg;
    if (events == null) return;

    var lastSentTime = DateTime.now();

    await for (final event in events) {
      final l2 = event[event.length >= 2 ? 1 : 0];
      for (var i = 0; i < l2.length; i++) {
        final isAttached = lastLeadState?.isAttached ?? true;

        final output = adaptiveHR.calculateHR(
          l2[i],
          isAttached,
        );

        final hr = output <= 0 ? null : output;
        if (hr == null &&
            DateTime.now().difference(lastSentTime) <
                const Duration(seconds: 2)) continue;

        yield hr;
        lastSentTime = DateTime.now();
      }
    }
    yield null;
  }

  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.convertedEcg;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;
    final callbacksTargets = [for (final callback in callbacks) callback.$2];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > maxBuffer) {
          channel.removeRange(0, channel.length - lengthBuffer);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          callback.call(channels, dataFilter);
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;

      x += sampleCount;
    }
  }

  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.bufferedEcg;
    if (events == null) return;

    await for (final event in events) {
      final isAttached = lastLeadState?.isAttached ?? true;
      yield udfFilter(
        channels: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
        dataFilter: dataFilter,
        isAttached: isAttached,
      );
    }
  }

  Stream<PlotChannels> _getPlot() async* {
    final events =
    isFiltered ? ble.streams.bufferedFilterEcg : ble.streams.bufferedEcg;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < event.list.length; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  Stream<PlotChannels?> _getPlotRT() async* {
    final events = ble.streams.plotEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % displayBuffer;
      final newXSkip = newX - newX;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      channelCount,
          (_) =>
      List<FlSpot>.generate(
        displayBuffer,
            (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        final index = skip(x + j.toDouble()).round();

        for (var k = 0; k < event.list.length; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }

        if (refreshBuffer > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBuffer).round();

          for (var k = 0; k < event.list.length; k++) {
            res.list[k][refreshedIndex] = FlSpot.nullSpot;
          }
        }
      }

      x += sampleCount;

      yield res.copyWith(
        pivot: (x % displayBuffer).toDouble(),
      );
    }
    yield null;
  }
}
