import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

class ModuleOxiBuiltin {
  static List<List<double>>? udfConvert({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    final res = List.generate(
      channelCount,
      (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.isEmpty) return null;

    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < channelCount; j++) {
        final msb = event[channelCount * 3 * i + j];
        final mid = event[channelCount * 3 * i + j + 1];
        final lsb = event[channelCount * 3 * i + j + 2];
        final value = toDouble(msb, mid, lsb);
        res[j][i] = value;
      }
    }
    return res;
  }
}

class ModuleOxi {
  ModuleOxi({
    required this.ble,
    this.commands = const Commands(),
    this.udfConvert = ModuleOxiBuiltin.udfConvert,
  });

  final ModuleBle ble;
  final Commands commands;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  /////

  final samplingRate = 100;
  int channelCount = 2;
  int sampleCount = 30;

  /// [maxBuffer] decides a maximum length of buffered Stream.
  /// For 40 seconds buffer and 100 sampling rate,
  /// It uses 40 * 100.
  int maxBuffer = 40 * 100;

  /// [lengthBuffer] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  /// For 30 seconds buffer and 100 sampling rate,
  /// It uses 30 * 100.
  int lengthBuffer = 30 * 100;

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  /// For 5 seconds buffer and 100 sampling rate,
  /// It uses 5 * 100 by default.
  int displayBuffer = 5 * 100;

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 100 ~/ 2 by default.
  int refreshBuffer = 100 ~/ 2;

  /// [isFiltered] decides whether pulse oxi signal is filtered.
  bool isFiltered = false;

  /// If [isCompressed] is `true`,
  /// pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter.initialize(8);

  AdaptiveHR adaptiveHR = AdaptiveHR(0);

  /// [callbacks] is functions to be executed periodically.
  final List<(void Function(Channels, DataFilter)? callback, int period)>
      callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedOxi]
  Channels? lastChannel;

  bool _isPaused = false;
  bool _isMeasuring = false;

  void initStream() {
    ble.streams
      ..rawOxi = null
      ..lzmaOxi = null
      ..convertedOxi = null
      ..eventOxi = null
      ..bufferedOxi = null
      ..bufferedFilterOxi = null
      ..plotOxi = null
      ..plotRTOxi = null;
  }

  /// [start] sends start_oxi_packet to device.
  Future<void> start() async {
    await _getConfig();
    await _sendStart();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams
      ..rawOxi = _getRaw().broadcast
      ..lzmaOxi = _getLzma().broadcast
      ..convertedOxi = _getConverted().broadcast
      ..eventOxi = _getEvent()?.broadcast
      ..bufferedOxi = _getBuffered().broadcast
      ..bufferedFilterOxi = _getBufferedFilter().broadcast
      ..plotOxi = _getPlot().broadcast
      ..plotRTOxi = _getPlotRT().broadcast;
  }

  /// [pause]
  Future<void> pause() async {
    _isPaused = !_isPaused;
  }

  /// [stop] sends stop_oxi_packet to device.
  Future<void> stop() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_OXI];
    await control.writePacket(commands.stopOxi);
  }

  /////////////////////////

  Future<void> _getConfig() async {
    final configuration = ble.points[Characteristics.CONFIG_OXI];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    channelCount = configurationValue[0];
    sampleCount = configurationValue[1];
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_OXI];
    await control.writePacket(commands.startEcg);
  }

  ////////////////////////////

  Stream<List<int>> _getRaw() async* {
    final measurement = ble.points[Characteristics.MEASUREMENT_OXI];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }

    await for (final res in measurement.lastValueStream) {
      if (!_isMeasuring || ble.device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.rawOxi;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed ? ble.streams.lzmaOxi : ble.streams.rawOxi;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  ///
  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.eventOxi;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  // TODO: implement filtering for pulse oxi
  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.convertedOxi;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;
    final callbacksTargets = [for (final callback in callbacks) callback.$2];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > maxBuffer) {
          channel.removeRange(0, channel.length - lengthBuffer);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          callback.call(channels, dataFilter);
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;

      x += sampleCount;
    }
  }

  // TODO: implement filtering for pulse oxi
  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.bufferedEcg;
    if (events == null) return;

    await for (final event in events) {
      yield event;
    }
  }

  Stream<PlotChannels> _getPlot() async* {
    final events =
        isFiltered ? ble.streams.bufferedFilterEcg : ble.streams.bufferedEcg;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < event.list.length; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  Stream<PlotChannels?> _getPlotRT() async* {
    final events = ble.streams.plotEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % displayBuffer;
      final newXSkip = newX - newX;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      channelCount,
      (_) => List<FlSpot>.generate(
        displayBuffer,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        final index = skip(x + j.toDouble()).round();

        for (var k = 0; k < event.list.length; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }

        if (refreshBuffer > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBuffer).round();

          for (var k = 0; k < event.list.length; k++) {
            res.list[k][refreshedIndex] = FlSpot.nullSpot;
          }
        }
      }

      x += sampleCount;

      yield res.copyWith(
        pivot: (x % displayBuffer).toDouble(),
      );
    }
    yield null;
  }
}
